

c***************************************************************
c***************************************************************
c***************************************************************
c
c       Zemlyanaya E.V 13.10.99
c       Solution of a single ODE equation 
c	  Numerov's approximation of the O(h^4) accuracy order
c
c		z''(x) + ff(x)*z'(x) + gg(x)*z(x) = uk(x)
c	
c         boundary conditions:
c		d1*z'(x_1) + f1*z(x_1) = g1
c		d2*z'(x_N) + f2*z(x_N) = g2
c
c		Interval:  x_1 =< x =< x_N
c
c INPUT:
c		N = size of dimensions;
c		h = stepsize of the finite-differense approximation;
c		ff(n),gg(n),uk(n) -- coefficients of equations; 
c		d1,d2,f1,f2,g1,g2 -- coefficients of boundary conditions;
c		av(n), bv(n) -- job dimensions (should be discribed in the main code)	
c
c
c OUTPUT:
c		z(n) -- soloution of the equation
c
c***************************************************************
c
      subroutine progon4(n,h,ff,gg,Uk,f1,f2,d1,d2,g1,g2,z,Av,bv)
      implicit real*8 (a-h,o-z)
      dimension ff(n),gg(n),Uk(n),z(n)
      DIMENSION av(n),bv(n),E(5,6),INDEX(5)

      N1=N-1
      N2=N-2
      N3=N-3
      N4=N-4
      N5=N-5
c
      C6  = 1.d0/6.d0
      C12 = 1.d0/12.d0
      H12 = h/12.d0
      H6  = h/6.d0
      hh  = h*h
      H1  = 1.d0/H
c
c---------------------------------- coeff. for z'' (i=2)
c
      vm1 = +1.d0-h12*(6.d0*ff(2)+5.d0*ff(3)+4.d0*ff(4)+
     ^				3.d0*ff(5))/12.d0
      vm2 = -2.d0-h12*(-2.d0*gg(2)*h+
     ^          (+20.d0*ff(2)-40.d0*ff(3)-24.d0*ff(4)-
     ^		     16.d0*ff(5))/12.d0)
      vm3 = +1.d0-h12*(+5.d0*gg(3)*h+
     ^          (-36.d0*ff(2)+72.d0*ff(4)+36.d0*ff(5))/12.d0)
      vm4 =    -h12*(-4.d0*gg(4)*h+
     ^          (+12.d0*ff(2)+40.d0*ff(3)-40.d0*ff(4)-
     ^		     48.d0*ff(5))/12.d0)
      vm5 =    -h12*(+gg(5)*h+
     ^         (-2.d0*ff(2)-5.d0*ff(3)-12.d0*ff(4)+25.d0*ff(5))/12.d0)
      vm6 =    -hh/12.d0*(2.d0*uk(2)-5.d0*uk(3)+4.d0*uk(4)-uk(5))

c      vl1 = -.5d0*h-hh/12.d0*(           -1.5d0*ff(2))
c      vl2 =      -hh/12.d0*(+3.d0*gg(2)*h+2.d0*ff(3))
c      vl3 = +.5d0*h-hh/12.d0*(-4.d0*gg(3)*h+1.5d0*ff(2)-.5d0*ff(4))
c      vl4 =      -hh/12.d0*(+gg(4)*h   -2.d0*ff(3))
c      vl5 =      -hh/12.d0*ff(4)*.5d0
c      vl6 =      -hh*h12*(-3.d0*uk(2)+4.d0*uk(3)-uk(4))
c
c---------------------------------- coeff. for z'' (i=n-1)
c
      vmn =  +1.d0-h12*(-6.d0*ff(n1)-5.d0*ff(n2)-4.d0*ff(n3)-
     ^			3.d0*ff(n4))/12.d0
      vmn1 = -2.d0-h12*(-2.d0*gg(n1)*h+
     ^          (-20.d0*ff(n1)+40.d0*ff(n2)+24.d0*ff(n3)+
     ^		     16.d0*ff(n4))/12.d0)
      vmn2 = +1.d0-h12*(+5.d0*gg(n2)*h+
     ^          (+36.d0*ff(n1)-72.d0*ff(n3)-36.d0*ff(n4))/12.d0)
      vmn3 =    -h12*(-4.d0*gg(n3)*h+
     ^          (-12.d0*ff(n1)-40.d0*ff(n2)+40.d0*ff(n3)+
     ^		     48.d0*ff(n4))/12.d0)
      vmn4 =    -h12*(+gg(n4)*h+
     ^          (+2.d0*ff(n1)+5.d0*ff(n2)+12.d0*ff(n3)-
     ^		     25.d0*ff(n4))/12.d0)
      vmn5 =    -hh/12.d0*(2.d0*uk(n1)-5.d0*uk(n2)+4.d0*uk(n3)-uk(n4))

c      vln  = +.5d0*h-hh/12.d0*(-1.5d0*ff(n1))
c      vln1 =    -hh/12.d0*(-3.d0*gg(n1)*h+2.d0 *ff(n2))
c      vln2 = -.5d0*h-hh/12.d0*(+4.d0*gg(n2)*h+1.5d0*ff(n1)-.5d0*ff(n3))
c      vln3 =    -hh/12.d0*(-gg(n3)*h   -2.d0 *ff(n2))
c      vln4 =    -hh/12.d0*ff(n3)*.5d0
c      vln5 =    -hh*h12*(+3.d0*uk(n1)-4.d0*uk(n2)+uk(n3))

	vl1 = -h12*3.d0
	vl2 = -h12*10.d0
	vl3 = +h12*18.d0
	vl4 = -h12*6.d0
	vl5 = +h12
	vl6 = 0.d0

	vln  = +h12*3.d0
	vln1 = +h12*10.d0
	vln2 = -h12*18.d0
	vln3 = +h12*6.d0
	vln4 = -h12
	vln5 = 0.d0

c
c---------------------------- coeff. for left sistem (i=1,2)
c
      a11 = -25.d0*d1+f1*h*12.d0
      a12 = +48.d0*d1
      a13 = -36.d0*d1
      a14 = 16.d0*d1
      a15 = -3.d0*d1
      cc1 = g1*h*12.d0
c
      a21 = vm1+FF(2)*vl1
      a22 = vm2+ff(2)*vl2+gg(2)*hh
      a23 = vm3+ff(2)*vl3
      a24 = vm4+ff(2)*vl4
      a25 = vm5+ff(2)*vl5
      cc2 = UK(2)*hh-vm6-vl6
c
c---------------------------------- coeff. for z' and z''(i=3)
c
         i = 3
       ip1 = i+1
       im1 = i-1
      am2  = 1.d0-((-ff(ip1)+3.d0*ff(im1)-2.d0*ff(i))*.5d0-
     ^				gg(im1)*h)*h12
     ^          +h*hh/24.d0*ff(i)*(+gg(im1)-(.5d0*ff(ip1)+
     ^		     1.5d0*ff(im1))/h)
      am4  = 1.d0-((-3.d0*ff(ip1)+ff(im1)+2.d0*ff(i))*.5d0-
     ^			gg(ip1)*h)*h12
     ^          +h*hh/24.d0*ff(i)*(-gg(ip1)-(1.5d0*ff(ip1)+
     ^		     .5d0*ff(im1))/h)
      am3  = -2.d0-((ff(ip1)-ff(im1))+gg(i)*h)*h/6.d0
     ^          +h*hh/12.d0*ff(i)*(ff(ip1)+ff(im1))/h
      s3   =  -(uk(ip1)+uk(im1)-2.d0*uk(i))*hh/12.d0
     ^          +hh*hh/24.d0*ff(i)*(uk(ip1)-uk(im1))

c
      al2  = -.5d0*h-hh/12.d0*(-(+ff(ip1)+3.d0*ff(im1))/2.d0+gg(im1)*h)
      al4  = +.5d0*h-hh/12.d0*(-(+ff(im1)+3.d0*ff(ip1))/2.d0-gg(ip1)*h)
      al3  = -hh/6.d0*(ff(ip1)+ff(im1))
      d3   = -hh*h12*(uk(ip1)-uk(im1))


c
c---------------------------------- coeff. for z' and z''(i=4)
c
       i   = 4
       ip1 = i+1
       im1 = i-1
      amm3  = 1.d0-((-ff(ip1)+3.d0*ff(im1)-2.d0*ff(i))*.5d0-
     ^			gg(im1)*h)*h12
     ^          +h*hh/24.d0*ff(i)*(+gg(im1)-(.5d0*ff(ip1)+
     ^		     1.5d0*ff(im1))/h)
      amm5 = 1.d0-((-3.d0*ff(ip1)+ff(im1)+2.d0*ff(i))*.5d0-
     ^			gg(ip1)*h)*h12
     ^          +h*hh/24.d0*ff(i)*(-gg(ip1)-(1.5d0*ff(ip1)+
     ^		     .5d0*ff(im1))/h)
      amm4 = -2.d0-((ff(ip1)-ff(im1))+gg(i)*h)*h/6.d0
     ^          +h*hh/12.d0*ff(i)*(ff(ip1)+ff(im1))/h
      s4   = -(uk(ip1)+uk(im1)-2.d0*uk(i))*hh/12.d0
     ^          +hh*hh/24.d0*ff(i)*(uk(ip1)-uk(im1))

c
      all3 = -.5d0*h-hh/12.d0*(-(+ff(ip1)+3.d0*ff(im1))/2.d0+gg(im1)*h)
      all5 = +.5d0*h-hh/12.d0*(-(+ff(im1)+3.d0*ff(ip1))/2.d0-gg(ip1)*h)
      all4 = -hh/6.d0*(ff(ip1)+ff(im1))
      d4   = -hh*h12*(uk(ip1)-uk(im1))


c
c---------------------------- coeff. for left sistem (i=3,4)
c
      a32 = am2+ff(3)*al2
      a33 = am3+ff(3)*al3 + gg(3)*hh
      a34 = am4+ff(3)*al4
      cc3 = UK(3)*hh-s3-ff(3)*d3
c
      a43 = amm3+ff(4)*all3
      a44 = amm4+ff(4)*all4 + gg(4)*hh
      a45 = amm5+ff(4)*all5
      cc4 = UK(4)*hh-s4-ff(4)*d4
c
c---------------------------- solution of the left sistem
c
        ap2 = a12*a21-a22*a11
        ap3 = a13*a21-a23*a11
        ap4 = a14*a21-a24*a11
        ap5 = a15*a21-a25*a11
        bet = cc1*a21-cc2*a11
c
        vp3 = ap3*a32-a33*ap2
        vp4 = ap4*a32-a34*ap2
        vp5 = ap5*a32
       vbet = bet*a32-cc3*ap2
c
       znam  = vp4*a43-a44*vp3
       av(5) = -(vp5*a43-a45*vp3)/znam
       bv(5) = (vbet*a43-cc4*vp3)/znam

c
c---------------------------------------------i=3,...,n-2
c
      DO  I=5,N2
c
      Im1 = I-1
      IP1 = I+1
c
      amim1= +1.d0-h12*((-ff(ip1)+3.d0*ff(im1)-2.d0*ff(i))/2.d0-
     ^			gg(im1)*h)
     ^          +h*hh/24.d0*ff(i)*(+gg(im1)-(.5d0*ff(ip1)+
     ^		     1.5d0*ff(im1))/h)
      amip1= +1.d0-h12*((-3.d0*ff(ip1)+ff(im1)+2.d0*ff(i))/2.d0-
     ^			gg(ip1)*h)
     ^          +h*hh/24.d0*ff(i)*(-gg(ip1)-(1.5d0*ff(ip1)+
     ^		     .5d0*ff(im1))/h)
      ami  = -2.d0-h6*(ff(ip1)-ff(im1)+gg(i)*h)
     ^          +h*hh/12.d0*ff(i)*(ff(ip1)+ff(im1))/h
      si   = -hh/12.d0*(uk(ip1)+uk(im1)-2.d0*uk(i))
     ^          +hh*hh/24.d0*ff(i)*(uk(ip1)-uk(im1))


c
      alim1  = -.5d0*h-hh/12.d0*(-(+ff(ip1)+3.d0*ff(im1))/2.d0+
     ^			gg(im1)*h)
      alip1  = +.5d0*h-hh/12.d0*(-(+ff(im1)+3.d0*ff(ip1))/2.d0-
     ^			gg(ip1)*h)
      ali    = -hh/6.d0*(ff(ip1)+ff(im1))
      di     = -hh*h12*(uk(ip1)-uk(im1))

c
      aim1 = amim1 + ff(i)*alim1
      aip1 = amip1 + ff(i)*alip1
      ai   = ami + ff(i)*ali+gg(i)*hh
      bi   = uk(i)*hh-ff(i)*di-si
c
      zna     = ai+aim1*av(i)
      Av(Ip1) = -aip1/zna
      Bv(Ip1) = (bi-aim1*bv(i))/zna
c
        end do
c
c------------------------------- SOLUTION OF THE RIGHT SISTEM
c
      DO 33 I = 1,5
      DO 33 J = 1,6
   33 E(I,J)  = 0.d0
c
      e(1,1) = d2*3.d0
      e(1,2) = -16.d0*d2
      E(1,3) = d2*36.d0
      E(1,4) = -48.d0*d2
      E(1,5) = f2*h*12.d0+25.d0*d2
      e(1,6) = g2*h*12.d0


c
      E(2,1) = vmn4+ff(n1)*vln4
      E(2,2) = vmn3+ff(n1)*vln3
      E(2,3) = vmn2+ff(n1)*vln2
      E(2,4) = vmn1+ff(n1)*vln1+gg(n1)*hh
      E(2,5) = vmn +ff(n1)*vln
      E(2,6) = Uk(n1)*hh-vmn5-vln5


c
      E(3,1) = 1.d0
      E(3,2) = -Av(N3)
      E(3,6) = Bv(N3)
c
      E(4,2) = 1.d0
      E(4,3) = -Av(N2)
      E(4,6) = Bv(N2)
c
      E(5,3) = 1.d0
      E(5,4) = -av(n1)
      E(5,6) = bv(n1)
c
      CALL MATIN2(E,5,5,MDIM,1,INDEX,NERROR,DETERM)
c
      Z(N)  = E(5,6)
      Z(N1) = E(4,6)
      Z(N2) = E(3,6)
      Z(N3) = E(2,6)
      Z(N4) = E(1,6)
c
C------------------------------  reverse moving
c
      DO  K = N5,4,-1
        kp1 = k+1
       Z(k) = Av(kp1)*Z(kp1)+bv(kp1)
        end do
c
      Z(3) = (vbet-vp5*Z(5)-vp4*Z(4))/vp3
      Z(2) = (bet-ap5*z(5)-ap4*Z(4)-ap3*Z(3))/ap2
      Z(1) = (cc1-a15*z(5)-a14*Z(4)-a13*Z(3)-a12*z(2))/a11
c
      RETURN
      END
c*
c******************************************************************
c*
c*
c******************************************************************
c*
      SUBROUTINE  MATIN2 (A,IDIM1,N1,IDIM2,N2,INDEX,NERROR,DETERM)
c     implicit real*8 (a-h,o-z)
C	 MATRIX INVERSION WITH ACCOMPANYING SOLUTION OF LINEAR EQUATIONS
      DOUBLE PRECISION A,DETERM,DETER,PIVOT,SWAP
      DIMENSION A(*),INDEX(*)
      DETER=1.0D0
      N=N1
      IEMAT=N+N2
      IDIM=IDIM1
      NMIN1=N-1
C	 THE ROUTINE DOES ITS OWN EVALUATION FOR DOUBLE SUBSCRIPTING OF
C	 ARRAY A.
      IPIVC=1-IDIM
C	 MAIN LOOP TO INVERT THE MATRIX
      DO 11 MAIN=1,N
      PIVOT=0.0D0
      IPIVC=IPIVC+IDIM
C	 SEARCH FOR NEXT PIVOT IN COLUMN MAIN.
      IPIVC1=IPIVC+MAIN-1
      IPIVC2=IPIVC +NMIN1
      DO 2 I1=IPIVC1,IPIVC2
      IF(DABS(A(I1))-DABS(PIVOT)) 2,2,1
    1 PIVOT=A(I1)
      LPIV=I1
    2 CONTINUE
C	 IS PIVOT DIFFERENT FROM ZERO
      IF(PIVOT) 3,15,3
C	 GET THE PIVOT-LINE INDICATOR AND SWAP LINES IF NECESSARY
    3 ICOL=LPIV-IPIVC+1
      INDEX(MAIN)=ICOL
      IF(ICOL-MAIN) 6,6,4
C	 COMPLEMENT THE DETERMINANT
    4 DETER=-DETER
C	 POINTER TO LINE PIVOT FOUND
      ICOL=ICOL-IDIM
C	 POINTER TO EXACT PIVOT LINE
      I3=MAIN-IDIM
      DO 5 I=1,IEMAT
      ICOL=ICOL+IDIM
      I3=I3+IDIM
      SWAP=A(I3)
      A(I3)=A(ICOL)
    5 A(ICOL)=SWAP
C	 COMPUTE DETERMINANT
    6 DETER=DETER*PIVOT
      PIVOT=1./PIVOT
C	 TRANSFORM PIVOT COLUMN
      I3=IPIVC+NMIN1
      DO 7 I=IPIVC,I3
    7 A(I)=-A(I)*PIVOT
      A(IPIVC1)=PIVOT
C	 PIVOT ELEMENT TRANSFORMED
C
C	 NOW CONVERT REST OF THE MATRIX
      I1=MAIN-IDIM
C	 POINTER TO PIVOT LINE ELEMENTS
      ICOL=1-IDIM
C	 GENERAL COLUMN POINTER
      DO 10 I=1,IEMAT
      ICOL=ICOL+IDIM
      I1=I1+IDIM
C	 POINTERS MOVED
      IF(I-MAIN) 8,10,8
C	 PIVOT COLUMN EXCLUDED
    8 JCOL=ICOL+NMIN1
      SWAP=A(I1)
      I3=IPIVC-1
      DO 9 I2=ICOL,JCOL
      I3=I3+1
    9 A(I2)=A(I2)+SWAP*A(I3)
      A(I1)=SWAP*PIVOT
   10 CONTINUE
   11 CONTINUE
C	 NOW REARRANGE THE MATRIX TO GET RIGHT INVERS
      DO 14 I1=1,N
      MAIN=N+1-I1
      LPIV=INDEX(MAIN)
      IF(LPIV-MAIN) 12,14,12
   12 ICOL=(LPIV-1)*IDIM+1
      JCOL=ICOL+NMIN1
      IPIVC=(MAIN-1)*IDIM+1-ICOL
      DO 13 I2=ICOL,JCOL
      I3=I2+IPIVC
      SWAP=A(I2)
      A(I2)=A(I3)
   13 A(I3)=SWAP
   14 CONTINUE
      DETERM=DETER
      NERROR=0
      RETURN
   15 NERROR=MAIN
      DETERM=DETER
      RETURN
      END
